# Rendered by CFEngine
#!/bin/bash

while true; do
  touch {{{vars.sys.workdir}}}/httpd/php/runalerts_all
 if [ -f {{{vars.sys.workdir}}}/httpd/php/runalerts_all ]; then
     {{{vars.sys.workdir}}}/httpd/php/bin/php {{{vars.cfe_internal_hub_vars.public_docroot}}}/index.php cli_tasks runalerts {{{vars.cfe_internal_php_runalerts.all[limit]}}} {{{vars.cfe_internal_php_runalerts.all[running]}}} all >/dev/null 2>&1
 fi
 sleep {{{vars.cfe_internal_php_runalerts.sleep_time}}}
done
